import { fileURLToPath } from 'url';
import { defu } from 'defu';
import { useLogger, resolveAlias, resolveFiles, defineNuxtModule, createResolver, addImportsDir, addTemplate, addServerHandler, updateTemplates } from '@nuxt/kit';
import inquirer from 'inquirer';
import { promises } from 'fs';
import fragmentImport from '@graphql-fragment-import/lib/inline-imports.js';
import { validateGraphQlDocuments } from '@graphql-tools/utils';
import { loadSchema } from '@graphql-tools/load';
import Table from 'cli-table';
import chalk from 'chalk';
import { concatAST, Source, parse } from 'graphql';
import { generate as generate$1, executeCodegen } from '@graphql-codegen/cli';
import * as PluginTypescript from '@graphql-codegen/typescript';
import * as PluginTypescriptOperations from '@graphql-codegen/typescript-operations';
import * as PluginSchemaAst from '@graphql-codegen/schema-ast';
import { oldVisit } from '@graphql-codegen/plugin-helpers';
import { pascalCase } from 'change-case-all';

const name = "nuxt-graphql-middleware";
const version = "3.0.1";

var GraphqlMiddlewareTemplate = /* @__PURE__ */ ((GraphqlMiddlewareTemplate2) => {
  GraphqlMiddlewareTemplate2["OperationTypes"] = "graphql-operations.d.ts";
  GraphqlMiddlewareTemplate2["ComposableContext"] = "nuxt-graphql-middleware.d.ts";
  GraphqlMiddlewareTemplate2["Documents"] = "graphql-documents.mjs";
  return GraphqlMiddlewareTemplate2;
})(GraphqlMiddlewareTemplate || {});

function falsy(value) {
  return value !== null && value !== void 0;
}

function getCodeResult(operations, typeName, serverApiPrefix) {
  const imports = [];
  let code = "";
  let nitroCode = "";
  const names = Object.keys(operations);
  if (names.length) {
    const lines = [];
    const nitroLines = [];
    names.forEach((name) => {
      const nameResult = pascalCase(name + typeName);
      imports.push(nameResult);
      const nameVariables = pascalCase(name + typeName + "Variables");
      const { hasVariables, variablesOptional } = operations[name];
      if (hasVariables) {
        imports.push(nameVariables);
      }
      const variablesType = hasVariables ? nameVariables : "null";
      lines.push(
        `    ${name}: [${variablesType}, ${variablesOptional ? "true" : "false"}, ${nameResult}]`
      );
      nitroLines.push(
        `    '${serverApiPrefix}/${typeName.toLowerCase()}/${name}': GraphqlMiddlewareResponse<Awaited<${nameResult}>>`
      );
    });
    code += `  export type GraphqlMiddleware${typeName} = {
${lines.join(",\n")}
  }
`;
    nitroCode += `${nitroLines.join("\n")}`;
  }
  return { code, imports, nitroCode };
}
const plugin$1 = (_schema, documents, config) => {
  const allAst = concatAST(documents.map((v) => v.document).filter(falsy));
  const operations = {
    query: {},
    mutation: {}
  };
  oldVisit(allAst, {
    enter: {
      OperationDefinition: (node) => {
        if ("name" in node && node.name?.value && "operation" in node && (node.operation === "query" || node.operation === "mutation")) {
          operations[node.operation][node.name.value] = {
            hasVariables: !!node.variableDefinitions?.length,
            variablesOptional: !!node.variableDefinitions?.every((v) => {
              return v.defaultValue;
            })
          };
        }
      }
    }
  });
  let code = "";
  let nitroCode = "";
  const imports = [];
  const resultQuery = getCodeResult(
    operations.query,
    "Query",
    config.serverApiPrefix
  );
  code += resultQuery.code;
  nitroCode += resultQuery.nitroCode;
  imports.push(...resultQuery.imports);
  const resultMutation = getCodeResult(
    operations.mutation,
    "Mutation",
    config.serverApiPrefix
  );
  code += "\n" + resultMutation.code;
  nitroCode += "\n" + resultMutation.nitroCode;
  imports.push(...resultMutation.imports);
  return `import {
  ${imports.join(",\n  ")}
} from './graphql-operations'


type GraphqlMiddlewareResponse<T> = {
  data: T
}

declare module '#build/nuxt-graphql-middleware' {
${code}
}

declare module 'nitropack' {
  type Awaited<T> = T extends PromiseLike<infer U> ? Awaited<U> : T
  interface InternalApi {
${nitroCode}
  }
}
`;
};

const PluginNuxtGraphqlMiddleware = {
  __proto__: null,
  plugin: plugin$1
};

const plugin = (_schema, documents, _config) => {
  const allAst = concatAST(documents.map((v) => v.document).filter(falsy));
  const operations = {
    query: {},
    mutation: {}
  };
  oldVisit(allAst, {
    enter: {
      OperationDefinition: (node) => {
        if (node.name?.value && node.loc?.source && (node.operation === "query" || node.operation === "mutation")) {
          operations[node.operation][node.name.value] = node.loc.source.body;
        }
      }
    }
  });
  return `const documents = ${JSON.stringify(operations, null, 2)};
export { documents };`;
};

const PluginNuxtGraphqlMiddlewareDocuments = {
  __proto__: null,
  plugin: plugin
};

function pluginLoader(name) {
  if (name === "@graphql-codegen/typescript") {
    return Promise.resolve(PluginTypescript);
  } else if (name === "@graphql-codegen/typescript-operations") {
    return Promise.resolve(PluginTypescriptOperations);
  } else if (name === "@graphql-codegen/typescript-nuxt-graphql-middleware") {
    return Promise.resolve(PluginNuxtGraphqlMiddleware);
  } else if (name === "@graphql-codegen/typescript-nuxt-graphql-middleware-documents") {
    return Promise.resolve(PluginNuxtGraphqlMiddlewareDocuments);
  } else {
    return Promise.resolve(PluginSchemaAst);
  }
}
function generateSchema(url, dest, writeToDisk) {
  return generate$1(
    {
      schema: url,
      pluginLoader,
      silent: true,
      errorsOnly: true,
      generates: {
        [dest]: {
          plugins: ["schema-ast"],
          config: {
            sort: true
          }
        }
      }
    },
    writeToDisk
  ).then((v) => v[0]);
}
function generateTemplates(documents, schemaPath, options) {
  return executeCodegen({
    schema: schemaPath,
    pluginLoader,
    silent: true,
    errorsOnly: true,
    documents,
    generates: {
      [GraphqlMiddlewareTemplate.OperationTypes]: {
        plugins: ["typescript", "typescript-operations"],
        config: options.codegenConfig
      },
      [GraphqlMiddlewareTemplate.ComposableContext]: {
        plugins: ["typescript-nuxt-graphql-middleware"],
        config: {
          serverApiPrefix: options.serverApiPrefix
        }
      },
      [GraphqlMiddlewareTemplate.Documents]: {
        plugins: ["typescript-nuxt-graphql-middleware-documents"]
      }
    }
  });
}

const logger = useLogger("nuxt-graphql-middleware");
const defaultOptions = {
  codegenConfig: {
    exportFragmentSpreadSubTypes: true,
    preResolveTypes: true,
    skipTypeNameForRoot: true,
    skipTypename: true,
    useTypeImports: true,
    onlyOperationTypes: true,
    namingConvention: {
      enumValues: "change-case-all#upperCaseFirst"
    }
  },
  downloadSchema: true,
  schemaPath: "./schema.graphql",
  serverApiPrefix: "/api/graphql_middleware",
  graphqlEndpoint: "",
  debug: false,
  includeComposables: true,
  documents: [],
  autoImportPatterns: ["**/*.{gql,graphql}", "!node_modules"]
};
function inlineFragments(source, resolver) {
  return fragmentImport(source, {
    resolveImport(identifier) {
      return resolver(identifier);
    },
    resolveOptions: {
      basedir: "./"
    }
  });
}
function validateOptions(options) {
  if (!options.graphqlEndpoint) {
    throw new Error("Missing graphqlEndpoint.");
  }
}
async function getSchemaPath(options, resolver, writeToDisk = false) {
  const dest = resolver(options.schemaPath);
  if (!options.downloadSchema) {
    const fileExists = await promises.access(dest).then(() => true).catch(() => false);
    if (!fileExists) {
      logger.error(
        '"downloadSchema" is set to false but no schema exists at ' + dest
      );
      throw new Error("Missing GraphQL schema.");
    }
    return dest;
  }
  if (!options.graphqlEndpoint) {
    throw new Error("Missing graphqlEndpoint config.");
  }
  const graphqlEndpoint = typeof options.graphqlEndpoint === "string" ? options.graphqlEndpoint : options.graphqlEndpoint();
  await generateSchema(graphqlEndpoint, dest, writeToDisk);
  return dest;
}
async function autoImportDocuments(patterns = [], srcResolver) {
  if (!patterns.length) {
    return Promise.resolve([]);
  }
  const files = (await resolveFiles(srcResolver(), patterns, {
    followSymbolicLinks: false
  })).filter((path) => {
    return !path.includes("schema.gql") && !path.includes("schema.graphql");
  });
  return Promise.all(
    files.map((filename) => {
      return promises.readFile(filename).then((v) => {
        const content = v.toString().trim();
        return {
          content,
          filename
        };
      });
    })
  );
}
function buildDocuments(providedDocuments = [], autoImportPatterns, resolver) {
  return autoImportDocuments(autoImportPatterns, resolver).then((importedDocuments) => {
    return [
      ...importedDocuments,
      ...providedDocuments.map((content) => {
        return {
          content,
          filename: "nuxt.config.ts"
        };
      })
    ].filter((v) => v.content);
  }).then((documents) => {
    return documents.map((v) => {
      try {
        return {
          content: inlineFragments(v.content, resolveAlias),
          filename: v.filename
        };
      } catch (e) {
        logger.error(
          "Failed to inline fragments for document: " + v.filename
        );
      }
      return null;
    }).filter(falsy);
  });
}
function parseDocument(document, srcDir) {
  let name = document.filename ? document.filename.replace(srcDir, "") : "";
  if (name.charAt(0) === "/") {
    name = name.slice(1);
  }
  const source = new Source(document.content, name);
  return parse(source);
}
function validateDocuments(schema, documents, srcDir) {
  for (let i = 0; i < documents.length; i++) {
    const document = documents[i];
    try {
      const node = parseDocument(document, srcDir);
      document.errors = validateGraphQlDocuments(schema, [
        node
      ]);
      const operation = node.definitions.find(
        (v) => v.kind === "OperationDefinition"
      );
      if (operation) {
        document.name = operation.name?.value;
        document.operation = operation.operation;
      }
      document.isValid = document.errors.length === 0;
    } catch (e) {
      document.errors = [e];
      document.isValid = false;
    }
  }
  return documents;
}
async function generate(options, schemaPath, resolver, srcDir, logEverything = false) {
  const schemaContent = await promises.readFile(schemaPath).then((v) => v.toString());
  const schema = await loadSchema(schemaContent, { loaders: [] });
  const documents = await buildDocuments(
    options.documents,
    options.autoImportPatterns,
    resolver
  );
  const validated = validateDocuments(schema, documents, srcDir);
  const templates = await generateTemplates(
    validated.filter((v) => v.isValid).map((v) => v.content),
    schemaPath,
    options
  );
  const hasErrors = validated.some((v) => !v.isValid);
  if (hasErrors || logEverything) {
    const table = new Table({
      head: ["Operation", "Name", "File", "Errors"].map((v) => chalk.white(v))
    });
    validated.forEach((document) => {
      if (logEverything || !document.isValid) {
        table.push(
          [
            document.operation || "",
            document.name || "",
            document.filename?.replace(srcDir, "") || "",
            document.errors?.join("\n\n") || ""
          ].map((v) => {
            if (document.isValid) {
              return v;
            }
            return chalk.red(v);
          })
        );
      }
    });
    logger.log(table.toString());
  }
  logger.info("Finished GraphQL code generation.");
  return { templates, hasErrors };
}

const module = defineNuxtModule({
  meta: {
    name,
    configKey: "graphqlMiddleware",
    version,
    compatibility: {
      nuxt: "^3.0.0"
    }
  },
  defaults: defaultOptions,
  async setup(passedOptions, nuxt) {
    const options = defu({}, passedOptions, defaultOptions);
    validateOptions(options);
    const rootDir = nuxt.options.rootDir;
    const moduleResolver = createResolver(import.meta.url).resolve;
    const srcDir = nuxt.options.srcDir;
    const srcResolver = createResolver(srcDir).resolve;
    const schemaPath = await getSchemaPath(
      options,
      srcResolver,
      options.downloadSchema
    );
    const runtimeDir = fileURLToPath(new URL("./runtime", import.meta.url));
    nuxt.options.build.transpile.push(runtimeDir);
    const ctx = {
      templates: []
    };
    let prompt = null;
    const generateHandler = async (isFirst = false) => {
      if (prompt && prompt.ui) {
        prompt.ui.close();
        prompt = null;
      }
      try {
        const { templates, hasErrors } = await generate(
          options,
          schemaPath,
          srcResolver,
          srcDir,
          isFirst
        );
        ctx.templates = templates;
        if (hasErrors) {
          throw new Error("Documents has errors.");
        }
      } catch (e) {
        console.log(e);
        logger.error("Failed to generate GraphQL files.");
        if (isFirst) {
          process.exit(1);
        }
        if (!options.downloadSchema) {
          return;
        }
        if (!nuxt.options.dev) {
          return;
        }
        process.stdout.write("\n");
        logger.restoreStd();
        prompt = inquirer.prompt({
          type: "confirm",
          name: "accept",
          message: "Do you want to reload the GraphQL schema?"
        }).then(async ({ accept }) => {
          if (accept) {
            await getSchemaPath(options, srcResolver);
            await generateHandler();
          }
        });
      }
    };
    await generateHandler(true);
    nuxt.options.runtimeConfig.public["nuxt-graphql-middleware"] = {
      serverApiPrefix: options.serverApiPrefix
    };
    nuxt.options.runtimeConfig.graphqlMiddleware = {
      rootDir
    };
    if (options.includeComposables) {
      addImportsDir(moduleResolver("runtime/composables"));
      nuxt.options.alias["#graphql-composable"] = moduleResolver(
        "runtime/composables"
      );
    }
    Object.values(GraphqlMiddlewareTemplate).forEach((filename) => {
      const result = addTemplate({
        write: true,
        filename,
        options: {
          nuxtGraphqlMiddleware: true
        },
        getContents: () => {
          return ctx.templates.find((v) => v.filename === filename)?.content || "";
        }
      });
      if (result.dst.includes(GraphqlMiddlewareTemplate.Documents)) {
        nuxt.options.alias["#graphql-documents"] = result.dst;
      } else if (result.dst.includes(GraphqlMiddlewareTemplate.OperationTypes)) {
        nuxt.options.alias["#graphql-operations"] = result.dst;
      }
    });
    addTemplate({
      write: true,
      filename: "graphql-documents.d.ts",
      getContents: () => {
        return `
import {
  GraphqlMiddlerwareQuery,
  GraphqlMiddlewareMutation,
} from '#build/nuxt-graphql-middleware'

declare module '#graphql-documents' {
  type Documents = {
    query: GraphqlMiddlerwareQuery
    mutation: GraphqlMiddlerwareMutation
  }
  const documents: Documents
  export { documents }
}
`;
      }
    });
    addServerHandler({
      handler: moduleResolver("./runtime/serverHandler/index"),
      route: options.serverApiPrefix + "/:operation/:name"
    });
    nuxt.hook("nitro:config", (nitroConfig) => {
      nitroConfig.externals = defu(
        typeof nitroConfig.externals === "object" ? nitroConfig.externals : {},
        {
          inline: [moduleResolver("./runtime")]
        }
      );
    });
    if (nuxt.options.dev) {
      nuxt.hook("nitro:build:before", (nitro) => {
        nuxt.hook("builder:watch", async (event, path) => {
          if (!path.match(/\.(gql|graphql)$/)) {
            return;
          }
          if (schemaPath.includes(path)) {
            return;
          }
          await generateHandler();
          await updateTemplates({
            filter: (template) => {
              return template.options && template.options.nuxtGraphqlMiddleware;
            }
          });
          await nitro.hooks.callHook("dev:reload");
        });
      });
    }
  }
});

export { module as default };
